/**
 * \file
 *
 * \brief Boot loader programming
 *
 * \author Atmel AVR32 Support \<avr32\@atmel.com\>
 *
 * Copyright (C) 2008-2010 Atmel Corporation. All rights reserved.
 *
 * \page License
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. The name of Atmel may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * 4. This software may only be redistributed and used in connection with an
 * Atmel AVR product.
 *
 * THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#include <errno.h>
#include <fcntl.h>
#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include <libmtd/mtd.h>

#include "upgrade.h"
#include "util.h"

static int write_bootloader(struct mtd *mtd, const void *buf, size_t len)
{
	struct mtd_sector	sector;
	unsigned int		offset;
	int			ret;

	/* Verify that all sectors are reachable and not bad */
	if (mtd_find_sector(mtd, 0, &sector)) {
		fprintf(stderr, "Failed to find sector at 0x%x: %s\n",
				0, strerror(errno));
		return 1;
	}

	while (1) {
		ret = mtd_sector_is_bad(mtd, &sector);
		if (ret < 0) {
			fprintf(stderr, "Bad block check failed at 0x%x: %s\n",
					sector.offset, strerror(errno));
			return 1;
		}
		if (ret) {
			fprintf(stderr, "Sector at 0x%x is bad\n",
					sector.offset);
			return 1;
		}

		if (sector.offset + sector.len >= len)
			break;

		if (mtd_next_sector(mtd, &sector)) {
			fprintf(stderr, "Failed to find sector at 0x%x: %s\n",
					sector.offset + sector.len,
					strerror(errno));
			return 1;
		}
	}

	/*
	 * This is the point of no return. The image is in RAM, and
	 * all flash sectors appear to be good. If something goes
	 * wrong now, there's not much we can do.
	 */
	fprintf(stderr, "Upgrading boot loader...");
	mtd_find_sector(mtd, 0, &sector);
	offset = 0;

	while (offset < len) {
		ret = mtd_erase_sector(mtd, &sector);
		if (ret < 0) {
			int err = errno;
			fprintf(stderr, "FAILED!\n");
			fprintf(stderr, "Failed to erase sector at 0x%x: %s\n",
					sector.offset, strerror(err));
			errno = err;
			return -1;
		}
		ret = mtd_write_sector(mtd, &sector,
				0, buf + offset, len - offset);
		if (ret < 0) {
			int err = errno;
			fprintf(stderr, "FAILED!\n");
			fprintf(stderr, "Failed to write sector at 0x%x: %s\n",
					sector.offset, strerror(err));
			errno = err;
			return -1;
		}

		offset += ret;
		if (offset < len)
			mtd_next_sector(mtd, &sector);
	}

	fprintf(stderr, "OK\n");

	return 0;
}

int upgrade_bootloader(const char *dev, const char *img)
{
	struct mtd	*mtd;
	void		*buf;
	int		fd;
	struct stat	sb;
	int		ret;

	fd = open(img, O_RDONLY);
	if (fd < 0) {
		fprintf(stderr, "Failed to open boot loader image `%s': %s\n",
				img, strerror(errno));
		goto err_open;
	}
	if (fstat(fd, &sb)) {
		perror("Failed to stat boot loader image");
		goto err_fstat;
	}

	mtd = mtd_open(dev);
	if (!mtd) {
		fprintf(stderr, "Failed to open MTD device %s: %s\n",
				dev, strerror(errno));
		goto err_mtd_open;
	}

	if (mtd_get_size(mtd) < sb.st_size) {
		fprintf(stderr, "U-Boot image is too large (%zu bytes)\n",
				sb.st_size);
		fprintf(stderr, "MTD partition size (%s): %u\n",
				dev, mtd_get_size(mtd));
		goto err_size;
	}

	/*
	 * Load the whole image into memory before proceeding -- we
	 * don't want to get caught halfway through with an unreadable
	 * image file.
	 */
	buf = malloc(sb.st_size);
	if (!buf) {
		fprintf(stderr,
			"Unable to allocate memory for boot loader image\n");
		goto err_malloc;
	}
	if (read_all(fd, buf, sb.st_size) != sb.st_size) {
		fprintf(stderr, "Unable to read boot loader image: %s\n",
				strerror(errno));
		goto err_read;
	}

	ret = write_bootloader(mtd, buf, sb.st_size);
	if (ret < 0) {
		fprintf(stderr, "Fatal error while upgrading boot loader.\n");
		fprintf(stderr, "Your system may be unbootable.\n");
	} else if (ret) {
		/* Error happened before touching the flash. */
		goto err_write;
	}

	free(buf);
	mtd_close(mtd);
	close(fd);

	return ret;

err_write:
err_read:
	free(buf);
err_malloc:
err_size:
	mtd_close(mtd);
err_mtd_open:
err_fstat:
	close(fd);
err_open:
	fprintf(stderr, "The boot loader will be left alone.\n");
	return 0;
}
