/**
 * \file
 *
 * \brief CRC32 Table
 *
 * This file was taken from the Linux MTD userspace project. It comes
 * without any license information, but the implementation is well known
 * and closely associated with the table in crc32.c. This file is
 * basically what you would get if you take any public domain version
 * found on the internet and apply our coding style to it.
 */
#ifndef CRC32_H
#define CRC32_H

#include <stdint.h>

extern const uint32_t crc32_table[256];

/**
 * \brief Return a 32-bit CRC of the contents of a buffer.
 *
 * To compute the CRC of a buffer split into multiple discontiguous
 * segments, simply call this function for each of the segment, and pass
 * the return value from the previous call as the \a val parameter.
 *
 * \param val Initial CRC value.
 * \param ss A pointer to be buffer.
 * \param len The length of the buffer in bytes.
 *
 * \return The updated CRC value after processing the buffer.
 */
static inline uint32_t
crc32(uint32_t val, const void *ss, int len)
{
	const unsigned char *s = ss;

	while (--len >= 0)
		val = crc32_table[(val ^ *s++) & 0xff] ^ (val >> 8);

	return val;
}

#endif
