/**
 * \file
 *
 * \brief Boot loader environment manipulation
 *
 * \author Atmel AVR32 Support \<avr32\@atmel.com\>
 *
 * Copyright (C) 2008-2010 Atmel Corporation. All rights reserved.
 *
 * \page License
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. The name of Atmel may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * 4. This software may only be redistributed and used in connection with an
 * Atmel AVR product.
 *
 * THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#define _GNU_SOURCE
#include <errno.h>
#include <malloc.h>
#include <stdbool.h>
#include <stdint.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>

#include "crc32.h"
#include "libmtd/mtd.h"
#include "upgrade.h"

struct environment {
	size_t		total_size;
	size_t		used_size;
	bool		is_redundant;
	char		*data;
};

static uint32_t env_crc(uint32_t crc, const void *buf, size_t len)
{
	crc ^= 0xffffffffL;
	crc = crc32(crc, buf, len);
	return crc ^ 0xffffffffL;
}

static size_t env_bytes_free(struct environment *env)
{
	return env->total_size - env->used_size - 1;
}

static uint32_t env_get_crc(struct environment *env)
{
	return *(uint32_t *)env->data;
}

static void env_set_crc(struct environment *env, uint32_t crc)
{
	*(uint32_t *)env->data = crc;
}

static bool env_is_obsolete(struct environment *env)
{
	if (!env->is_redundant)
		return false;

	return *(uint8_t *)(env->data + 4) != 1;
}

static void env_set_obsolete(struct environment *env, bool is_obsolete)
{
	if (env->is_redundant)
		*(uint8_t *)(env->data + 4) = is_obsolete ? 0 : 1;
}

static char *env_get_data(struct environment *env, size_t *size)
{
	if (env->is_redundant) {
		if (size)
			*size = env->total_size - 5;
		return env->data + 5;
	} else {
		if (size)
			*size = env->total_size - 4;
		return env->data + 4;
	}
}

struct environment *env_create(size_t size, bool redundant)
{
	struct environment	*env;
	char			*data;

	env = malloc(sizeof(struct environment));
	if (!env)
		return NULL;

	env->total_size = size;
	env->used_size = 4;
	if (redundant)
		env->used_size++;
	env->is_redundant = redundant;
	env->data = malloc(size);
	if (!env->data) {
		free(env);
		return NULL;
	}
	memset(env->data, 0xff, size);

	data = env_get_data(env, NULL);
	data[0] = data[1] = 0;

	return env;
}

void env_destroy(struct environment *env)
{
	free(env->data);
	free(env);
}

const char *env_get(struct environment *env, const char *name)
{
	size_t		name_len;
	size_t		data_size;
	size_t		i;
	char		*data;

	data = env_get_data(env, &data_size);
	name_len = strlen(name);

	for (i = 0; i < data_size && data[i]; ) {
		if (data[i + name_len] == '='
				&& memcmp(data + i, name, name_len) == 0)
			return data + i + name_len + 1;
		i += strnlen(data + i, data_size - i) + 1;
	}

	return NULL;
}

static void env_remove_var(struct environment *env, char *var, size_t len)
{
	memmove(var, var + len, env->used_size - (var - env->data));
	env->used_size -= len;
	memset(env->data + env->used_size + 1, 0xff, len);
	env->data[env->used_size] = 0;
}

int env_set(struct environment *env, const char *name, const char *value)
{
	size_t		name_len;
	size_t		var_len;
	size_t		data_size;
	size_t		i;
	char		*data;

	data = env_get_data(env, &data_size);
	name_len = strlen(name);

	for (i = 0; i < data_size && data[i]; ) {
		var_len = strnlen(data + i, data_size - i) + 1;
		if (var_len > data_size) {
			errno = EFAULT;
			return -1;
		}
		if (data[i + name_len] == '='
				&& memcmp(data + i, name, name_len) == 0) {
			/* Variable already exists. Delete it. */
			env_remove_var(env, data + i, var_len);
			break;
		}
		i += strnlen(data + i, data_size - i) + 1;
	}

	/*
	 * Append the new variable to the end if there's enough free
	 * space.
	 */
	var_len = name_len + strlen(value) + 2;
	if (var_len > env_bytes_free(env)) {
		errno = ENOSPC;
		return -1;
	}

	sprintf(env->data + env->used_size, "%s=%s", name, value);
	env->used_size += var_len;
	env->data[env->used_size] = 0;

	return 0;
}

static int env_read_mtd_internal(struct environment *env, struct mtd *mtd)
{
	size_t		data_size;
	size_t		i;
	uint32_t	crc;
	char		*data;

	if (mtd_read(mtd, 0, env->data, env->total_size) < 0)
		return -1;

	crc = env_get_crc(env);
	data = env_get_data(env, &data_size);

	if (env_crc(0, data, data_size) != crc) {
		errno = EILSEQ;
		return -1;
	}

	/* Find the end marker */
	data = env_get_data(env, &data_size);
	for (i = 1; i < data_size; i++) {
		if (data[i - 1] == 0 && data[i] == 0) {
			env->used_size = i - 1;
			return 0;
		}
	}

	/* No end marker found -- invalid */
	errno = EINVAL;
	return -1;
}

int env_read_from_mtd(struct environment *env, struct mtd *main_mtd,
		struct mtd *backup_mtd)
{
	if ((env->is_redundant && !backup_mtd)
			|| (!env->is_redundant && backup_mtd)) {
		errno = EINVAL;
		return -1;
	}

	if (backup_mtd) {
		if (env_read_mtd_internal(env, main_mtd) < 0
				|| env_is_obsolete(env)) {
			if (env_read_mtd_internal(env, backup_mtd) < 0
					|| env_is_obsolete(env)) {
				errno = EIO;
				return -1;
			}
		}
	} else {
		if (env_read_mtd_internal(env, main_mtd) < 0) {
			errno = EIO;
			return -1;
		}
	}

	return 0;
}

void env_refresh(struct environment *env)
{
	size_t	data_size;
	char	*data;

	env->data[env->used_size] = 0;
	memset(env->data + env->used_size + 1, 0xff,
			env->total_size - env->used_size - 1);
	data = env_get_data(env, &data_size);
	env_set_crc(env, env_crc(0, data, data_size));
	env_set_obsolete(env, false);
}

int env_write_to_mtd(struct environment *env, struct mtd *mtd)
{
	struct mtd_sector	sector;
	size_t			data_offset = 0;
	size_t			used_size;
	mtd_size_t		mtd_size;
	int			ret;

	mtd_size = mtd_get_size(mtd);

	/* Add terminating '\0' */
	used_size = env->used_size + 1;

	for (ret = mtd_find_sector(mtd, 0, &sector);
			ret >= 0; ret = mtd_next_sector(mtd, &sector)) {
		if (mtd_sector_is_bad(mtd, &sector))
			continue;

		if (mtd_erase_sector(mtd, &sector) < 0)
			return -1;

		if (data_offset < used_size) {
			ret = mtd_write_sector(mtd, &sector, 0,
					env->data + data_offset,
					used_size - data_offset);
			if (ret < 0)
				return -1;
			data_offset += ret;
		}

		if (sector.offset + sector.len >= mtd_size)
			break;
	}

	if (data_offset < used_size) {
		errno = ENOSPC;
		return -1;
	}

	return 0;
}

int env_write_to_file(struct environment *env, int fd)
{
	size_t	written = 0;
	ssize_t	ret;

	while (written < env->total_size) {
		ret = write(fd, env->data + written, env->total_size - written);
		if (ret < 0)
			return ret;
		written += ret;
	}

	return written;
}
