/**
 * \file
 *
 * \brief Journaling Flash Filesystem 2
 *
 * \author Atmel AVR32 Support \<avr32\@atmel.com\>
 *
 * Copyright (C) 2008-2010 Atmel Corporation. All rights reserved.
 *
 * \page License
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. The name of Atmel may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * 4. This software may only be redistributed and used in connection with an
 * Atmel AVR product.
 *
 * THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#define _XOPEN_SOURCE 500
#include <errno.h>
#include <linux/jffs2.h>
#include <mtd/mtd-user.h>
#include <malloc.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <libmtd/mtd.h>
#include <libmtd/mtd_internal.h>

#include "crc32.h"
#include "upgrade.h"
#include "filesystem.h"

#define JINT16(x)	({ jint16_t _ret; _ret.v16 = (x); _ret; })
#define JINT32(x)	({ jint32_t _ret; _ret.v32 = (x); _ret; })

static inline jint32_t cpu_to_jint32(uint32_t x)
{
	jint32_t ret;
	ret.v32 = x;
	return ret;
}

static void jffs2_init_cleanmarker(struct jffs2_unknown_node *cleanmarker)
{
	uint32_t crc;

	memset(cleanmarker, 0, sizeof(*cleanmarker));
	cleanmarker->magic = JINT16(JFFS2_MAGIC_BITMASK);
	cleanmarker->nodetype = JINT16(JFFS2_NODETYPE_CLEANMARKER);
	cleanmarker->totlen = JINT32(sizeof(*cleanmarker));

	crc = crc32(0, cleanmarker,
			offsetof(struct jffs2_unknown_node, hdr_crc));
	cleanmarker->hdr_crc = cpu_to_jint32(crc);
}

static int jffs2_write_inline_cleanmarker(struct mtd *mtd,
		const struct mtd_sector *sector)
{
	struct jffs2_unknown_node *cleanmarker = mtd->volume_data;

	if (!cleanmarker) {
		cleanmarker = malloc(sizeof(*cleanmarker));
		if (!cleanmarker)
			return -1;
		mtd->volume_data = cleanmarker;
		jffs2_init_cleanmarker(cleanmarker);
	}

	if (pwrite(mtd->fd, cleanmarker, sizeof(*cleanmarker), sector->offset)
			!= sizeof(*cleanmarker))
		return -1;

	return 0;
}

static int jffs2_write_cleanmarker(struct mtd *mtd,
		const struct mtd_sector *sector)
{
	if (mtd->info.type == MTD_NORFLASH)
		return jffs2_write_inline_cleanmarker(mtd, sector);

	/*
	 * NAND flash and DataFlash is too difficult for us to handle;
	 * let the kernel figure it out on the first mount.
	 */
	return 0;
}

int fs_init_jffs2(const char *dev, const char *mountpoint)
{
	struct mtd_sector	sector;
	mtd_size_t		part_size;
	struct mtd		*mtd;
	mtd_off_t		offset;
	const char		*what;
	int			err;
	int			ret;

	mtd = mtd_open(dev);
	if (!mtd) {
		fprintf(stderr, "Failed to open MTD device %s: %s\n",
				dev, strerror(errno));
		return -1;
	}

	/*
	 * Make sure there's no stale filesystem data hanging around.
	 * This might cause us to leak a small amount of memory, but
	 * probably not, and it's better than crashing.
	 */
	mtd->volume_data = NULL;

	part_size = mtd_get_size(mtd);
	ret = mtd_find_sector(mtd, 0, &sector);

	for (offset = 0; offset < part_size; offset += sector.len,
			ret = mtd_next_sector(mtd, &sector)) {
		if (ret < 0) {
			what = "Could not find sector";
			goto fail;
		}

		fprintf(stderr, "\rFormatting %s as JFFS2: %u/%u bytes",
			dev, offset, part_size);

		ret = mtd_sector_is_bad(mtd, &sector);
		if (ret < 0) {
			what = "Bad block check failed";
			goto fail;
		}
		if (ret) {
			fprintf(stderr, ", skipped bad block @ 0x%x\n",
					sector.offset);
			continue;
		}

		ret = mtd_erase_sector(mtd, &sector);
		if (ret < 0) {
			what = "Failed to erase sector";
			goto fail;
		}

		ret = jffs2_write_cleanmarker(mtd, &sector);
		if (ret < 0) {
			what = "Failed to write cleanmarker";
			goto fail;
		}
	}

	fprintf(stderr, " - OK\n");

	mtd_close(mtd);

	return fs_mount(dev, mountpoint, "jffs2");

fail:
	err = errno;
	fprintf(stderr, " - FAILED!\n");
	fprintf(stderr, "%s at 0x%x: %s\n", what, offset, strerror(err));
	mtd_close(mtd);
	return ret;
}
