/**
 * \file
 *
 * \brief libmtd public interface
 *
 * \author Atmel AVR32 Support \<avr32\@atmel.com\>
 *
 * Copyright (C) 2008-2010 Atmel Corporation. All rights reserved.
 *
 * \page License
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. The name of Atmel may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * 4. This software may only be redistributed and used in connection with an
 * Atmel AVR product.
 *
 * THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#ifndef LIBMTD_MTD_H_INCLUDED
#define LIBMTD_MTD_H_INCLUDED

#include <fcntl.h>
#include <stdbool.h>
#include <stdint.h>

/**
 * MTD offset type. This type is used to represent byte offsets into
 * MTD devices.
 */
typedef uint32_t mtd_off_t;

/**
 * MTD size type. This type is used to represent byte sizes associated
 * with MTD devices.
 */
typedef uint32_t mtd_size_t;

struct mtd;

/**
 * A MTD sector. This structure describes one of the sectors which
 * makes up a MTD device. A sector is the smallest unit which can be
 * erased.
 */
struct mtd_sector {
	/** Byte offset into the MTD device where this sector starts */
	mtd_off_t	offset;
	/** Length of this sector in bytes */
	mtd_size_t	len;
	/** For internal use by the MTD library */
	void		*mtd_priv;
};

int mtd_get_devname(char *buf, size_t buf_len, const char *name);

struct mtd *mtd_open_device(const char *devname);
struct mtd *mtd_open(const char *name);
void mtd_close(struct mtd *mtd);

mtd_size_t mtd_get_size(const struct mtd *mtd);

int mtd_find_sector(const struct mtd *mtd, mtd_off_t offset,
		struct mtd_sector *sector);
int mtd_next_sector(const struct mtd *mtd, struct mtd_sector *sector);
int mtd_sector_is_bad(struct mtd *mtd, const struct mtd_sector *sector);

int mtd_erase_sector(const struct mtd *mtd, const struct mtd_sector *sector);
ssize_t mtd_write_sector(const struct mtd *mtd, const struct mtd_sector *sector,
		mtd_off_t offset, const void *buf, mtd_size_t len);

ssize_t mtd_read(const struct mtd *mtd, mtd_off_t offset, void *buf,
		mtd_size_t len);

#endif /* LIBMTD_MTD_H_INCLUDED */
