/**
 * \file
 *
 * \brief Locating the next sector
 *
 * \author Atmel AVR32 Support \<avr32\@atmel.com\>
 *
 * Copyright (C) 2008-2010 Atmel Corporation. All rights reserved.
 *
 * \page License
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. The name of Atmel may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * 4. This software may only be redistributed and used in connection with an
 * Atmel AVR product.
 *
 * THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#include <mtd/mtd-user.h>

#include "mtd.h"
#include "mtd_internal.h"

/**
 * Find the sector immediately following a given sector.
 *
 * @param[in] mtd MTD device.
 * @param[in, out] sector A sector in a MTD device. On success, this
 * is updated to describe the sector immediately following it.
 *
 * @retval 0 The next sector was found, and @a sector was initialized
 * with its offset and length.
 * @retval -1 No next sector was found, and errno was set to ENXIO.
 * This can happen if @a sector describes the last sector in the
 * device, or if @a sector was invalid to begin with.
 */
int __export mtd_next_sector(const struct mtd *mtd, struct mtd_sector *sector)
{
	struct region_info_user	*region = sector->mtd_priv;
	mtd_off_t		offset;
	mtd_size_t		region_end;

	region_end = region->offset + region->erasesize * region->numblocks;
	offset = sector->offset + region->erasesize;
	if (offset < region_end) {
		sector->offset = offset;
		return 0;
	}

	region = _mtd_find_region(mtd, offset);
	if (!region)
		return -1;

	sector->len = region->erasesize;
	sector->mtd_priv = region;

	return 0;
}
