/**
 * \file
 *
 * \brief Obtaining a reference to a MTD device
 *
 * \author Atmel AVR32 Support \<avr32\@atmel.com\>
 *
 * Copyright (C) 2008-2010 Atmel Corporation. All rights reserved.
 *
 * \page License
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. The name of Atmel may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * 4. This software may only be redistributed and used in connection with an
 * Atmel AVR product.
 *
 * THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#define _GNU_SOURCE
#include <errno.h>
#include <fcntl.h>
#include <malloc.h>
#include <mtd/mtd-user.h>
#include <stdio.h>
#include <string.h>
#include <sys/ioctl.h>
#include <unistd.h>

#include "mtd.h"
#include "mtd_internal.h"

/**
 * Close a MTD device. This will release all resources associated with
 * a MTD device and free the MTD device object.
 *
 * @param[in] mtd MTD device to be released.
 */
void __export mtd_close(struct mtd *mtd)
{
	close(mtd->fd);
	free(mtd->region);
	free(mtd);
}

/**
 * Open a MTD device.
 *
 * @param[in] devname The name of the MTD device to open. This must be
 * specified as "mtdX" where X is an unsigned number corresponding to
 * one of the /dev/mtdX special files.
 *
 * @return A pointer to an object representing the MTD device, or NULL
 * if it could not be opened.
 */
__export struct mtd *mtd_open_device(const char *devname)
{
	char		*filename;
	struct mtd	*mtd = NULL;
	int		errno_saved;
	int		fd;
	int		ret;
	int		i;

	mtd = malloc(sizeof(struct mtd));
	if (!mtd)
		return NULL;
	memset(mtd, 0, sizeof(*mtd));

	if (asprintf(&filename, "/dev/%s", devname) < 0) {
		free(mtd);
		return NULL;
	}

	fd = open(filename, O_RDWR);
	free(filename);
	if (fd < 0) {
		free(mtd);
		return NULL;
	}

	mtd->fd = fd;

	ret = ioctl(fd, MEMGETINFO, &mtd->info);
	if (ret)
		goto fail;

	ret = ioctl(fd, MEMGETREGIONCOUNT, &mtd->region_count);
	if (ret)
		goto fail;

	if (!mtd->region_count) {
		/*
		 * No region info: create a fake erase region spanning
		 * the whole device
		 */
		mtd->region = malloc(sizeof(mtd->region[0]));
		if (!mtd->region)
			goto fail;
		memset(mtd->region, 0, sizeof(mtd->region[0]));
		mtd->region[0].erasesize = mtd->info.erasesize;
		mtd->region[0].numblocks = mtd->info.size / mtd->info.erasesize;
		mtd->region_count = 1;
	} else {
		unsigned long alloc_len;

		alloc_len = mtd->region_count * sizeof(mtd->region[0]);
		mtd->region = malloc(alloc_len);
		if (!mtd->region)
			goto fail;
		memset(mtd->region, 0, alloc_len);

		for (i = 0; i < mtd->region_count; i++) {
			mtd->region[i].regionindex = i;
			ret = ioctl(fd, MEMGETREGIONINFO, &mtd->region[i]);
			if (ret)
				goto fail;
			mtd->sector_count += mtd->region[i].numblocks;
		}
	}
	
	/*
	 * Assume we need to check for bad blocks initially. Set to
	 * false if MEMGETBADBLOCK fails with EOPNOTSUPP and the
	 * device is not NAND flash.
	 */
	mtd->check_bad_blocks = true;

	return mtd;

fail:
	errno_saved = errno;
	mtd_close(mtd);
	errno = errno_saved;
	return NULL;
}

