/**
 * \file
 *
 * \brief Checking for bad sectors
 *
 * \author Atmel AVR32 Support \<avr32\@atmel.com\>
 *
 * Copyright (C) 2008-2010 Atmel Corporation. All rights reserved.
 *
 * \page License
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. The name of Atmel may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * 4. This software may only be redistributed and used in connection with an
 * Atmel AVR product.
 *
 * THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#include <errno.h>
#include <mtd/mtd-user.h>
#include <sys/ioctl.h>
#include <sys/types.h>

#include "mtd_internal.h"
#include "mtd.h"

/**
 * Check if a MTD sector is bad. On devices which can have bad sectors
 * (e.g. NAND flash), this will check if a given sector is bad. On
 * other devices, it will always return 0.
 *
 * @param[in] mtd MTD device.
 * @param[in] sector The sector to check.
 *
 * @retval 0 The sector is good.
 * @retval 1 The sector is bad.
 * @retval -1 It could not be determined whether the sector is good or
 * bad.
 */
int __export mtd_sector_is_bad(struct mtd *mtd, const struct mtd_sector *sector)
{
	if (mtd->check_bad_blocks) {
		loff_t	offset = sector->offset;
		int	ret;

		ret = ioctl(mtd->fd, MEMGETBADBLOCK, &offset);
		if (ret < 0) {
			if (errno == EOPNOTSUPP
					&& mtd->info.type != MTD_NANDFLASH)
				mtd->check_bad_blocks = false;
			else
				return -1;
		} else if (ret > 0) {
			return 1;
		}
	}

	return 0;
}
