/**
 * \file
 *
 * \brief AVR32 Buildroot Upgrade Utility
 *
 * \author Atmel AVR32 Support \<avr32\@atmel.com\>
 *
 * Copyright (C) 2010 Atmel Corporation. All rights reserved.
 *
 * \page License
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. The name of Atmel may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * 4. This software may only be redistributed and used in connection with an
 * Atmel AVR product.
 *
 * THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "libmtd/mtd.h"
#include "board.h"
#include "upgrade.h"
#include "util.h"

#define __unused		__attribute__((__unused__))

/*
 * Assume /boot is on a separate partition by default. If not, board.h
 * needs to define this as "/boot/uImage".
 */
#ifndef BOARD_UIMAGE_PATH
# define BOARD_UIMAGE_PATH	"/uImage"
#endif

#define TARGET_DIR		"/target"
#define IMAGE_DIR		"/mnt/" BOARD_IMAGE_DIR

#define BOOTLOADER_IMAGE	IMAGE_DIR "/u-boot.bin"
#define FILE_SYSTEM_IMAGE	IMAGE_DIR "/rootfs.tar"
#define ENV_BACKUP_FILE		IMAGE_DIR "/env-backup.bin"

__unused static int upgrade_environment(const char *dev)
{
	struct mtd		*mtd;
	struct environment	*old;
	struct environment	*new;
	int			fd;
	int			ret = -1;
	int			i;

	mtd = mtd_open(dev);
	if (!mtd) {
		fprintf(stderr, "Failed to open env device %s: %s\n",
				dev, strerror(errno));
		return -1;
	}

	old = env_create(mtd_get_size(mtd), false);
	if (!old) {
		perror("Failed to create buffer for old env");
		goto out_close_mtd;
	}

	if (env_read_from_mtd(old, mtd, NULL) < 0) {
		perror("Failed to read old env");
		goto out_free_old;
	}

	rename(ENV_BACKUP_FILE, ENV_BACKUP_FILE ".old");

	fd = open(ENV_BACKUP_FILE, O_WRONLY | O_CREAT | O_TRUNC, 0666);
	if (fd < 0) {
		fprintf(stderr, "Failed to open env backup file %s: %s\n",
				ENV_BACKUP_FILE, strerror(errno));
		goto out_free_old;
	}

	if (env_write_to_file(old, fd) < 0) {
		fprintf(stderr, "Failed to write old environment to %s: %s\n",
				ENV_BACKUP_FILE, strerror(errno));
		goto out_close_fd;
	}

	fprintf(stderr, "Backup of old environment stored in %s\n",
			ENV_BACKUP_FILE);

	new = env_create(mtd_get_size(mtd), false);
	if (!new) {
		perror("Failed to create buffer for new env");
		goto out_close_fd;
	}

	for (i = 0; i < ARRAY_LEN(board_env_preserve); i++) {
		const char *value;

		value = env_get(old, board_env_preserve[i]);
		if (value)
			env_set(new, board_env_preserve[i], value);
	}

	env_set(new, "bootcmd_mtd", "setenv bootargs $(bootargs) "
			BOARD_MTD_BOOTARGS ";"
			" fsload 10400000 " BOARD_UIMAGE_PATH "; bootm");
	env_set(new, "bootcmd_nfs", "dhcp; setenv bootargs $(bootargs)"
			" root=/dev/nfs ip=$(ipaddr):$(serverip)"
			":$(gatewayip):$(netmask):$(hostname):eth0:off"
			" nfsroot=$(serverip):$(rootpath); bootm");
	env_set(new, "upgrade", "setenv bootargs 'mtdparts="
			BOARD_UPGRADE_MTDPARTS "'; mmc init;"
			" fatload mmc 0:1 10400000 "
			BOARD_IMAGE_DIR "/uImage-upgrade; bootm");
	env_set(new, "bootcmd", "run bootcmd_mtd");

	env_refresh(new);

	if (env_write_to_mtd(new, mtd) < 0) {
		/*
		 * Failed to update the environment -- let's try to
		 * put the old one back before we fail...
		 */
		perror("Failed to write new environment");
		if (env_write_to_mtd(old, mtd) < 0)
			perror("Failed to restore old environment");
		goto out_free_new;
	}

	ret = 0;

out_free_new:
	env_destroy(new);
out_close_fd:
	close(fd);
out_free_old:
	env_destroy(old);
out_close_mtd:
	mtd_close(mtd);

	return ret;
}

int main(int argc, char *argv[])
{
	int ret;

	/* Don't make files writeable for group and others */
	umask(S_IWGRP | S_IWOTH);
	/* We want to handle pipe write errors synchronously */
	signal(SIGPIPE, SIG_IGN);

#ifdef CONFIG_UPGRADE_BOOTLOADER
	/* Upgrade bootloader first */
	ret = upgrade_bootloader(BOARD_BOOTLOADER_DEVICE, BOOTLOADER_IMAGE);
	if (ret < 0)
		return 1;
#endif

	ret = board_mount_filesystems(TARGET_DIR);
	if (ret < 0)
		return 3;

	ret = unpack_filesystem(TARGET_DIR, FILE_SYSTEM_IMAGE);
	if (ret && errno == ENOENT)
		ret = unpack_filesystem(TARGET_DIR, FILE_SYSTEM_IMAGE ".bz2");
	if (ret && errno == ENOENT)
		ret = unpack_filesystem(TARGET_DIR, FILE_SYSTEM_IMAGE ".gz");
	if (ret)
		fprintf(stderr, "Failed to unpack file system image\n");

	if (!file_exists(TARGET_DIR "/boot/uImage")) {
		if (file_exists(IMAGE_DIR "/uImage")) {
			/*
			 * Ignore errors from the mkdir command; if the
			 * directory doesn't exist, we'll get an error
			 * from the copy command as well.
			 */
			mkdir_p(TARGET_DIR "/boot", 0777);
			if (copy_file(IMAGE_DIR "/uImage",
						TARGET_DIR "/boot/uImage") < 0)
				fprintf(stderr, "ERROR: Failed to copy "
						"uImage into /boot\n");
		} else {
			fprintf(stderr, "WARNING: /boot/uImage not present\n");
		}
	}

	board_unmount_filesystems(TARGET_DIR);

#ifdef CONFIG_UPGRADE_ENV
	if (!ret)
		ret = upgrade_environment(BOARD_ENV_DEVICE);
#endif

	fprintf(stderr, "Syncing filesystems...\n");
	sync();

	if (ret)
		return 10;

	return 0;
}
