/**
 * \file
 *
 * \brief Spawning subprocesses
 *
 * \author Atmel AVR32 Support \<avr32\@atmel.com\>
 *
 * Copyright (C) 2010 Atmel Corporation. All rights reserved.
 *
 * \page License
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. The name of Atmel may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * 4. This software may only be redistributed and used in connection with an
 * Atmel AVR product.
 *
 * THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#include "debug.h"
#include <errno.h>
#include <malloc.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include "subprocess.h"

/**
 * \brief Run helper program as a subprocess and wait for completion.
 *
 * \param filename Full path to the helper program.
 * \param ... Zero or more positional arguments for the helper program.
 * \return Exit status from the program, or -1 if something went wrong
 * on the way.
 */
int run_helper(const char *filename, ...)
{
	pid_t childpid;
	char *argv[MAX_NR_ARGS + 1];
	int i = 0;
	int ret = -1;
	va_list ap;

	argv[i++] = strdup(filename);
	pr_debug("Running helper: %s", filename);
	va_start(ap, filename);
	while (i < MAX_NR_ARGS) {
		const char *arg = va_arg(ap, const char *);

		if (!arg)
			break;
		pr_debug(" %s", arg);
		argv[i++] = strdup(arg);
	}
	argv[i] = NULL;
	va_end(ap);
	pr_debug("\n");

	childpid = fork();
	if (childpid < 0) {
		perror("Failed to create subprocess");
		goto out;
	} else if (childpid == 0) {
		execv(argv[0], argv);
		fprintf(stderr, "Failed to execute %s: %s\n",
				argv[0], strerror(errno));
		_exit(1);
	} else {
		int status;

		if (waitpid(childpid, &status, 0) != childpid) {
			perror("waitpid() failed");
			goto out;
		}
		if (!WIFEXITED(status) || WEXITSTATUS(status) != 0) {
			if (WIFEXITED(status))
				fprintf(stderr, "Helper program %s terminated"
						" abnormally with status %d\n",
						argv[0], WEXITSTATUS(status));
			else if (WIFSIGNALED(status))
				fprintf(stderr, "Helper program %s killed"
						" by signal %d\n",
						argv[0], WTERMSIG(status));
			else
				fprintf(stderr, "Helper program %s died for"
						" unknown reasons: %d\n",
						argv[0], status);
			ret = status;
		} else {
			/* Success */
			ret = 0;
		}
	}

out:
	while (i--)
		free(argv[i]);

	return ret;
}
