/**
 * \file
 *
 * \brief Unpacking target images
 *
 * \author Atmel AVR32 Support \<avr32\@atmel.com\>
 *
 * Copyright (C) 2008-2010 Atmel Corporation. All rights reserved.
 *
 * \page License
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. The name of Atmel may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * 4. This software may only be redistributed and used in connection with an
 * Atmel AVR product.
 *
 * THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#define _GNU_SOURCE
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include "upgrade.h"

#define TAR_EXECUTABLE		"/bin/tar"
#define GUNZIP_EXECUTABLE	"/bin/gunzip"
#define BUNZIP2_EXECUTABLE	"/usr/bin/bunzip2"
#define CHUNK_SIZE		65536

enum image_fmt {
	IMAGE_FMT_TAR,
	IMAGE_FMT_TAR_GZ,
	IMAGE_FMT_TAR_BZ2,
};

static int spawn_decompressor(enum image_fmt fmt, int *tar_pipe,
		int image_fd, pid_t *decomp_pid)
{
	int	decomp_pipe[2];
	pid_t	pid;

	if (pipe(decomp_pipe) < 0) {
		perror("Failed to create decompressor pipe");
		goto err_pipe;
	}

	pid = fork();
	if (pid < 0) {
		perror("Failed to create decompressor process");
		goto err_fork;
	}

	if (pid == 0) {
		/*
		 * This is the child. Close the write side and
		 * the filesystem image fd, and point stdin to
		 * the read side.
		 */
		close(decomp_pipe[1]);
		close(image_fd);
		dup2(decomp_pipe[0], STDIN_FILENO);
		close(decomp_pipe[0]);

		/* Feed the write side of the tar pipe from stdout */
		close(tar_pipe[0]);
		dup2(tar_pipe[1], STDOUT_FILENO);
		close(tar_pipe[1]);

		if (fmt == IMAGE_FMT_TAR_BZ2) {
			execl(BUNZIP2_EXECUTABLE,
					BUNZIP2_EXECUTABLE,
					"-c", NULL);
			fprintf(stderr, "Failed to execute %s: %s\n",
					BUNZIP2_EXECUTABLE,
					strerror(errno));
		} else {
			execl(GUNZIP_EXECUTABLE,
					GUNZIP_EXECUTABLE,
					"-c", NULL);
			fprintf(stderr, "Failed to execute %s: %s\n",
					GUNZIP_EXECUTABLE,
					strerror(errno));
		}
		_exit(1);
	}

	*decomp_pid = pid;

	close(decomp_pipe[0]);
	return decomp_pipe[1];

err_fork:
	close(decomp_pipe[0]);
	close(decomp_pipe[1]);

err_pipe:
	return -1;
}

static void kill_decompressor(pid_t pid)
{
	int retries = 10;

	if (kill(pid, SIGINT))
		return;

	while (retries-- > 0) {
		if (waitpid(pid, NULL, WNOHANG) != 0)
			return;
		usleep(50000);
	}

	if (kill(pid, SIGKILL))
		return;

	waitpid(pid, NULL, 0);
}

static int wait_for_child(pid_t pid, const char *what)
{
	int		status;

	if (waitpid(pid, &status, 0) < 0) {
		fprintf(stderr, "Failed to wait for %s: %s", what,
				strerror(errno));
		return -1;
	} else if (WIFEXITED(status)) {
		if (WEXITSTATUS(status) != 0) {
			fprintf(stderr, "%s exited with status %d\n", what,
					WEXITSTATUS(status));
			return -1;
		}
	} else {
		fprintf(stderr, "%s terminated abnormally (%d)\n",
				what, status);
		return -1;
	}

	return 0;
}

int unpack_filesystem(const char *rootpath, const char *image)
{
	long		bytes_xfered;
	long		last_xfered;
	long		ret;
	int		tar_pipe[2];
	int		imagefd;
	int		writefd;
	pid_t		decomp_pid;
	pid_t		unpack_pid;
	size_t		img_name_len;
	enum image_fmt	fmt = IMAGE_FMT_TAR;
	struct stat	sb;

	img_name_len = strlen(image);
	if (img_name_len >= 7
			&& memcmp(image + img_name_len - 7, ".tar.gz", 7) == 0)
		fmt = IMAGE_FMT_TAR_GZ;
	else if (img_name_len >= 8
			&& memcmp(image + img_name_len - 8, ".tar.bz2", 8) == 0)
		fmt = IMAGE_FMT_TAR_BZ2;

	imagefd = open(image, O_RDONLY);
	if (imagefd < 0) {
		fprintf(stderr, "Failed to open filesystem image %s: %s\n",
				image, strerror(errno));
		return -1;
	}
	if (fstat(imagefd, &sb)) {
		perror("Failed to stat filesystem image");
		goto err_fstat;
	}

	if (pipe(tar_pipe) < 0) {
		perror("Failed to create pipe");
		goto err_pipe;
	}

	writefd = tar_pipe[1];

	/* First, create the decompressor process */
	decomp_pid = -1;
	if (fmt != IMAGE_FMT_TAR) {
		int fd;

		fd = spawn_decompressor(fmt, tar_pipe, imagefd, &decomp_pid);
		if (fd < 0)
			goto err_decomp;

		close(writefd);
		writefd = fd;
	}

	/* Then, create the unpacking process */
	unpack_pid = fork();
	if (unpack_pid < 0) {
		perror("Failed to create unpacker process");
		goto err_fork_unpack;
	}

	if (unpack_pid == 0) {
		/*
		 * Hook up the read side of tar_pipefd to stdin and
		 * close the rest.
		 */
		close(writefd);
		close(imagefd);
		dup2(tar_pipe[0], STDIN_FILENO);
		close(tar_pipe[0]);

		if (chdir(rootpath) < 0) {
			fprintf(stderr, "Failed to enter directory %s: %s\n",
					rootpath, strerror(errno));
			_exit(1);
		}
		execl(TAR_EXECUTABLE, TAR_EXECUTABLE, "x", NULL);
		fprintf(stderr, "Failed to execute %s: %s\n", TAR_EXECUTABLE,
				strerror(errno));
		_exit(1);
	}

	/* Close the read side and start feeding the pipe */
	close(tar_pipe[0]);
	ret = 0;
	last_xfered = -1024 * 1024;
	for (bytes_xfered = 0; bytes_xfered < sb.st_size; bytes_xfered += ret) {
		if (bytes_xfered - last_xfered > 1024 * 512) {
			fprintf(stderr,
				"\rUnpacking root filesystem: %lu/%lu bytes",
				bytes_xfered, sb.st_size);
			last_xfered = bytes_xfered;
		}
		ret = splice(imagefd, NULL, writefd, NULL, CHUNK_SIZE, 0);
		if (ret <= 0)
			break;
	}

	if (ret < 0) {
		perror("\nsplice() failed");
	} else if (bytes_xfered != sb.st_size) {
		fprintf(stderr, "\nPremature end of file\n");
		ret = -1;
	} else {
		fprintf(stderr, " - done\n");
		ret = 0;
	}
	close(writefd);

	if (decomp_pid > 0) {
		int tmp = wait_for_child(decomp_pid, "decompressor");
		if (!ret)
			ret = tmp;
	}
	if (unpack_pid > 0) {
		int tmp = wait_for_child(unpack_pid, "unpacker");
		if (!ret)
			ret = tmp;
	}

	close(imagefd);
	return ret;

err_fork_unpack:
	if (decomp_pid > 0)
		kill_decompressor(decomp_pid);
err_decomp:
	close(writefd);
	close(tar_pipe[0]);
err_pipe:
err_fstat:
	close(imagefd);
	return -1;
}
