/**
 * \file
 *
 * \brief Misc utility functions
 *
 * \author Atmel AVR32 Support \<avr32\@atmel.com\>
 *
 * Copyright (C) 2008-2010 Atmel Corporation. All rights reserved.
 *
 * \page License
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. The name of Atmel may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * 4. This software may only be redistributed and used in connection with an
 * Atmel AVR product.
 *
 * THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#include <errno.h>
#include <fcntl.h>
#include <malloc.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "util.h"

ssize_t read_all(int fd, void *buf, size_t count)
{
	ssize_t	ret;
	size_t	nread = 0;

	while (nread < count) {
		ret = read(fd, buf + nread, count - nread);
		if (ret < 0)
			return ret;
		nread += ret;
	}

	return nread;
}

int mkdir_p(const char *path, mode_t mode)
{
	struct stat	sb;
	int		ret;

	if (!stat(path, &sb)) {
		if (S_ISDIR(sb.st_mode)) {
			return 0;
		} else {
			errno = ENOTDIR;
			return 0;
		}
	} else if (errno == ENOENT) {
		char *parent = strdup(path);
		char *slash = strrchr(parent, '/');

		ret = 0;
		if (slash && slash != parent) {
			*slash = '\0';
			ret = mkdir_p(parent, mode);
		}
		free(parent);
		if (ret)
			return ret;
		ret = mkdir(path, mode);
	} else {
		ret = -1;
	}

	return ret;
}

bool file_exists(const char *path)
{
	struct stat sb;

	if (lstat(path, &sb) < 0)
		return false;
	else
		return true;
}

int copy_file(const char *src_path, const char *dst_path)
{
	int src_fd, dst_fd;
	int ret = -1;
	char buf[65536];

	src_fd = open(src_path, O_RDONLY);
	if (src_fd < 0) {
		fprintf(stderr, "Failed to open %s for reading: %s\n",
				src_path, strerror(errno));
		goto err_open_src;
	}
	dst_fd = open(dst_path, O_WRONLY | O_CREAT | O_TRUNC, 0666);
	if (dst_fd < 0) {
		fprintf(stderr, "Failed to open %s for writing: %s\n",
				dst_path, strerror(errno));
		goto err_open_dst;
	}

	while (1) {
		ssize_t nread, nwritten;

		nread = read(src_fd, buf, sizeof(buf));
		if (nread < 0) {
			perror("Read error");
			goto err_read;
		}
		if (nread == 0)
			break;

		for (nwritten = 0; nwritten < nread; ) {
			ssize_t tmp;

			tmp = write(dst_fd, buf + nwritten, nread - nwritten);
			if (tmp < 0) {
				perror("Write error");
				goto err_write;
			}
			nwritten += tmp;
		}
	}

	ret = 0;

err_write:
err_read:
	close(dst_fd);
err_open_dst:
	close(src_fd);
err_open_src:
	return ret;
}
